@@ -1,4 +1,11 @@
 #
+# bootmisc.sh	Miscellaneous things to be done during bootup.
+#
+# Version:	@(#)bootmisc.sh  1.10  22-Jun-1998  miquels@cistron.nl
+#
+
+. /etc/default/rcS
+#
 # Put a nologin file in /etc to prevent people from logging in before
 # system startup is complete.
 #
@@ -21,6 +28,7 @@ fi
   ! \( -name quota.user -uid 0 \) \
   ! \( -name quota.group -uid 0 \) \
     -depth -exec rm -rf -- {} \; )
+rm -f /tmp/.X*-lock
 #
 # Clean up any stale locks.
 #
@@ -29,7 +37,7 @@ fi
 #
 # Clean up /var/run and create /var/run/utmp so that we can login.
 #
-[ "$VERBOSE" != no ] && echo -n "/var/run "
+[ "$VERBOSE" != no ] && echo -n "/var/run"
 ( cd /var/run && \
 	find . ! -type d ! -name utmp ! -name innd.pid ! -name random-seed \
 	-exec rm -f -- {} \; )
