@@ -5,7 +5,7 @@
  *		     init [0123456SsQqAaBbCc]
  *		  telinit [0123456SsQqAaBbCc]
  *
- * Version:	@(#)init.c  2.73  03-Jan-1998  MvS
+ * Version:	@(#)init.c  2.75-2  02-Jun-1998  miquels@cistron.nl
  *
  *		This file is part of the sysvinit suite,
  *		Copyright 1991-1998 Miquel van Smoorenburg.
@@ -14,6 +14,15 @@
  *		modify it under the terms of the GNU General Public License
  *		as published by the Free Software Foundation; either version
  *		2 of the License, or (at your option) any later version.
+ *
+ * Modified:	21 Feb 1998, Al Viro:
+ *		'U' flag added to telinit. It forces init to re-exec itself
+ *		(passing its state through exec, certainly).
+ *		May be useful for smoother (heh) upgrades.
+ *		24 Feb 1998, AV:
+ *		did_boot made global and added to state - thanks, Miquel.
+ *		Yet another file descriptors leak - close state pipe if 
+ *		re_exec fails.
  */
 
 #include <sys/types.h>
@@ -42,8 +51,8 @@
 #include "reboot.h"
 #include "set.h"
 
-#if !defined(__GLIBC__)
-#  define CLEAN_UTMP 1
+#ifndef SIGPWR
+#  define SIGPWR SIGUSR2
 #endif
 
 /* Set a signal handler. */
@@ -52,9 +61,9 @@
 				sigaction(sig, &sa, NULL);
 
 /* Version information */
-char *Version = "@(#) init 2.73 03-Jan-1998 MvS";
-char *bootmsg = "version 2.73 booting";
-#define E_VERSION "INIT_VERSION=sysvinit-2.73"
+char *Version = "@(#) init 2.75-2  02-Jun-1998  miquels@cistron.nl";
+char *bootmsg = "version 2.75 booting";
+#define E_VERSION "INIT_VERSION=sysvinit-2.75"
 
 CHILD *family = NULL;		/* The linked list of all entries */
 CHILD *newFamily = NULL;	/* The list after inittab re-read */
@@ -84,6 +93,13 @@ struct utmp utproto;		/* Only used for sizeof(utproto.ut_id) */
 char *console_dev = CONSOLE;	/* Console device. */
 int wrote_iosave = 0;		/* Did we write ioctl.save yet? */
 int pipe_fd = -1;		/* /dev/initctl */
+int did_boot = 0;		/* Did we already do BOOT* stuff? */
+
+/*	Used by re-exec part */
+int reload = 0;			/* Should we do initialization stuff? */
+char *myname="/sbin/init";	/* What should we exec */
+int oops_error;			/* Used by some of the re-exec code. */
+char *Signature = "12567362";	/* Signature for re-exec fd */
 
 /* Macro to see if this is a special action */
 #define ISPOWER(i) ((i) == POWERWAIT || (i) == POWERFAIL || \
@@ -114,199 +130,40 @@ struct actions {
 };
 
 /*
- *	Set the process title. We do not check for overflow of
- *	the stack space since we know there is plenty for
- *	our needs and we'll never use more than 10 bytes anyway.
+ *	State parser token table (see receive_state)
  */
-int setproctitle(char *fmt, ...)
-{
-  va_list ap;
-  int len;
-  char buf[256];
-
-  buf[0] = 0;
-
-  va_start(ap, fmt);
-  len = vsprintf(buf, fmt, ap);
-  va_end(ap);
-
-  memset(argv0, 0, maxproclen + 1);
-  strncpy(argv0, buf, maxproclen);
-
-  return len;
-}
-
-/* Open a device with retries. */
-int serial_open(char *dev, int mode)
-{
-  int f, fd;
-  int m;
-
-  /* Open device in nonblocking mode. */
-  m = mode | O_NONBLOCK;
-
-  /* Retry the open five times. */
-  for(f = 0;f < 5; f++)
-	if ((fd = open(dev, m)) >= 0) break;
-  if (fd < 0) return(fd);
-
-  /* Set original flags. */
-  if (m != mode)
-  	fcntl(fd, F_SETFL, mode);
-  return(fd);
-}
-
-/* We got a signal (HUP PWR WINCH ALRM INT) */
-void signal_handler(int sig)
-{
-  ADDSET(got_signals, sig);
-}
-
-/* SIGCHLD: one of our children has died. */
-void chld_handler()
-{
-  CHILD *ch;
-  int pid, st;
-
-  /* Find out which process(es) this was (were) */
-  while((pid = waitpid(-1, &st, WNOHANG)) != 0) {
-	if (errno == ECHILD) break;
-#ifdef __alpha__
-	/* Some (old) Linux/Alpha version have a buggy library */
-	if (pid == -ECHILD) break;
-#endif
-	for( ch = family; ch; ch = ch->next )
-		if ( ch->pid == pid && (ch->flags & RUNNING) ) {
-#if DEBUG
-			log(L_VB, "chld_handler: marked %d as zombie", ch->pid);
-#endif
-			ADDSET(got_signals, SIGCHLD);
-			ch->exstat = st;
-			ch->flags |= ZOMBIE;
-			if (ch->new) {
-				ch->new->exstat = st;
-				ch->new->flags |= ZOMBIE;
-			}
-			break;
-		}
-#if DEBUG
-	if (ch == NULL)
-		log(L_VB, "chld_handler: unknown child %d exited.", pid);
-#endif
-  }
-}
-
-/*
- * Linux ignores all signals sent to init when the
- * SIG_DFL handler is installed. Therefore we must catch SIGTSTP
- * and SIGCONT, or else they won't work....
- *
- * The SIGCONT handler
- */
-void cont_handler()
-{
-  got_cont = 1;
-}
-
-/*
- * OOPS: segmentation violation!
- */
-void segv_handler()
-{
-  log(L_VB, "PANIC: segmentation violation! giving up..");
-  while(1) pause();
-}
-
-/*
- * The SIGSTOP & SIGTSTP handler
- */
-void stop_handler()
-{
-  got_cont = 0;
-  while(!got_cont) pause();
-  got_cont = 0;
-}
-
-/* Set terminal settings to reasonable defaults */
-void set_term(int how)
-{
-  struct termios tty;
-  int fd, fd2;
-  int restore_ok = 0;
-
-  if ((fd = serial_open(console_dev, O_RDWR|O_NOCTTY)) < 0) {
-	log(L_VB, "can't open %s", console_dev);
-	return;
-  }
-
-  /* Do we want to save or restore modes. */
-  if (how == 0) {
-	/* Restore. */
-	if (wrote_iosave && (fd2 = open(IOSAVE, O_RDONLY)) >= 0) {
-		if (read(fd2, &tty, sizeof(tty)) == sizeof(tty))
-			restore_ok = 1;
-		close(fd2);
-	}
-	if (restore_ok == 0) {
-		/* Get old settings */
-		(void) ioctl(fd, TCGETS, &tty);
-
-		/* Set speed etc. */
-		tty.c_cflag &= CBAUD|CBAUDEX|CSIZE|CSTOPB|PARENB|PARODD;
-		tty.c_cflag |= HUPCL|CLOCAL;
-
-		/* Set the most important characters */
-		tty.c_cc[VINTR]  = 3;
-		tty.c_cc[VQUIT]  = 28;
-		tty.c_cc[VERASE] = 127;
-		tty.c_cc[VKILL]  = 24;
-		tty.c_cc[VEOF]   = 4;
-		tty.c_cc[VTIME]  = 0;
-		tty.c_cc[VMIN]   = 1;
-		tty.c_cc[VSTART] = 17;
-		tty.c_cc[VSTOP]  = 19;
-		tty.c_cc[VSUSP]  = 26;
-	}
-
-	/* Set pre and post processing */
-	tty.c_iflag = IGNPAR|ICRNL|IXON|IXANY;
-	tty.c_oflag = OPOST|ONLCR;
-	tty.c_lflag = ISIG|ICANON|ECHO|ECHOCTL|ECHOPRT|ECHOKE;
-
-	/* Now set the terminal line. */
-	(void) ioctl(fd, TCSETS, &tty);
-	(void) close(fd);
-	return;
-  }
-
-  /* Save the terminal settings. */
-  (void) ioctl(fd, TCGETS, &tty);
-  if ((fd2 = open(IOSAVE, O_WRONLY|O_CREAT|O_TRUNC, 0600)) < 0) {
-	if (errno != EROFS)
-		log(L_VB, "can't open(%s, O_WRONLY): %s",
-			IOSAVE, strerror(errno));
-	(void) close(fd);
-	(void) close(fd2);
-	return;
-  }
-  (void) write(fd2, &tty, sizeof(struct termios));
-  wrote_iosave = 1;
-  (void) close(fd);
-  (void) close(fd2);
-}
-
-/*
- *	Print to the system console
- */
-void print(char *s)
-{
-	int fd;
-
-	if ((fd = serial_open(console_dev, O_WRONLY|O_NOCTTY|O_NDELAY)) >= 0) {
-		write(fd, s, strlen(s));
-		close(fd);
-	}
-}
+struct {
+  char name[4];	
+  int cmd;
+} cmds[] = {
+  { "VER", 	   C_VER	},
+  { "END",	   C_END	},
+  { "REC",	   C_REC	},
+  { "EOR",	   C_EOR	},
+  { "LEV",	   C_LEV	},
+  { "FL ",	   C_FLAG	},
+  { "AC ",	   C_ACTION	},
+  { "CMD",	   C_PROCESS	},
+  { "PID",	   C_PID	},
+  { "EXS",	   C_EXS	},
+  { "-RL",	   D_RUNLEVEL	},
+  { "-TL",	   D_THISLEVEL	},
+  { "-PL",	   D_PREVLEVEL	},
+  { "-SI",	   D_GOTSIGN	},
+  { "-WR",	   D_WROTE_REBOOT},
+  { "-ST",	   D_SLTIME	},
+  { "-DB",	   D_DIDBOOT	},
+  { "",	   	   0		}
+};
+struct {
+	char *name;
+	int mask;
+} flags[]={
+	{"RU",RUNNING},
+	{"DE",DEMAND},
+	{"XD",XECUTED},
+	{NULL,0}
+};
 
 /*
  *	Sleep a number of seconds
@@ -340,31 +197,461 @@ void do_sleep(int sec)
 		alarm(old_alrm - sec > 0 ? old_alrm - sec : 1);
 }
 
+/*
+ *	Send the state info of the previous running init to
+ *	the new one, in a version-independant way.
+ */
+void send_state(int fd)
+{
+	FILE	*fp;
+	CHILD	*p;
+	int	i,val;
+
+	fp = fdopen(fd,"w");
+
+	fprintf(fp, "VER%s\n", Version);
+	fprintf(fp, "-RL%c\n", runlevel);
+	fprintf(fp, "-TL%c\n", thislevel);
+	fprintf(fp, "-PL%c\n", prevlevel);
+	fprintf(fp, "-SI%u\n", got_signals);
+	fprintf(fp, "-WR%d\n", wrote_reboot);
+	fprintf(fp, "-ST%d\n", sltime);
+	fprintf(fp, "-DB%d\n", did_boot);
+
+	for (p = family; p; p = p->next) {
+		fprintf(fp, "REC%s\n", p->id);
+		fprintf(fp, "LEV%s\n", p->rlevel);
+		for (i = 0, val = p->flags; flags[i].mask; i++)
+			if (val & flags[i].mask) {
+				val &= ~flags[i].mask;
+				fprintf(fp, "FL %s\n",flags[i].name);
+			}
+		fprintf(fp, "PID%d\n",p->pid);
+		fprintf(fp, "EXS%u\n",p->exstat);
+		for(i = 0; actions[i].act; i++)
+			if (actions[i].act == p->action) {
+				fprintf(fp, "AC %s\n", actions[i].name);
+				break;
+			}
+		fprintf(fp, "CMD%s\n", p->process);
+		fprintf(fp, "EOR\n");
+	}
+	fprintf(fp, "END\n");
+	fclose(fp);
+}
+
+/*
+ *	Read a string from a file descriptor.
+ *	FIXME: why not use fgets() ?
+ */
+static int get_string(char *p, int size, FILE *f)
+{
+	int	c;
+
+	while ((c = getc(f)) != EOF && c != '\n') {
+		if (--size > 0)
+			*p++ = c;
+	}
+	*p = '\0';
+	return (c != EOF) && (size > 0);
+}
+
+/*
+ *	Read trailing data from the state pipe until we see a newline.
+ */
+static int get_void(FILE *f)
+{
+	int	c;
+
+	while ((c = getc(f)) != EOF && c != '\n')
+		;
+
+	return (c != EOF);
+}
+
+/*
+ *	Read the next "command" from the state pipe.
+ */
+static int get_cmd(FILE *f)
+{
+	char	cmd[4] = "   ";
+	int	i;
+
+	if (fread(cmd, 1, 3, f) != 3)
+		return C_EOF;
+
+	for(i = 0; cmds[i].cmd && strcmp(cmds[i].name, cmd) != 0; i++)
+		;
+	return cmds[i].cmd;
+}
+
+/*
+ *	Read a CHILD * from the state pipe.
+ */
+static CHILD *get_record(FILE *f)
+{
+	int	cmd;
+	char	s[32];
+	int	i;
+	CHILD	*p;
+
+	do {
+		switch (cmd = get_cmd(f)) {
+			case C_END:
+				get_void(f);
+				return NULL;
+			case 0:
+				get_void(f);
+				break;
+			case C_REC:
+				break;
+			case D_RUNLEVEL:
+				fscanf(f, "%c\n", &runlevel);
+				break;
+			case D_THISLEVEL:
+				fscanf(f, "%c\n", &thislevel);
+				break;
+			case D_PREVLEVEL:
+				fscanf(f, "%c\n", &prevlevel);
+				break;
+			case D_GOTSIGN:
+				fscanf(f, "%u\n", &got_signals);
+				break;
+			case D_WROTE_REBOOT:
+				fscanf(f, "%d\n", &wrote_reboot);
+				break;
+			case D_SLTIME:
+				fscanf(f, "%d\n", &sltime);
+				break;
+			case D_DIDBOOT:
+				fscanf(f, "%d\n", &did_boot);
+				break;
+			default:
+				if (cmd > 0 || cmd == C_EOF) {
+					oops_error = -1;
+					return NULL;
+				}
+		}
+	} while (cmd != C_REC);
+
+	while ((p = (CHILD *)malloc(sizeof(CHILD))) == NULL ) {
+		log(L_VB, "out of memory");
+		do_sleep(5);
+	}
+	memset(p, 0, sizeof(CHILD));
+	get_string(p->id, sizeof(p->id), f);
+	do switch(cmd = get_cmd(f)) {
+		case 0:
+		case C_EOR:
+			get_void(f);
+			break;
+		case C_PID:
+			fscanf(f, "%d\n", &(p->pid));
+			break;
+		case C_EXS:
+			fscanf(f, "%u\n", &(p->exstat));
+			break;
+		case C_LEV:
+			get_string(p->rlevel, sizeof(p->rlevel), f);
+			break;
+		case C_PROCESS:
+			get_string(p->process, sizeof(p->process), f);
+			break;
+		case C_FLAG:
+			get_string(s, 32, f);
+			for(i = 0; flags[i].name; i++) {
+				if (strcmp(flags[i].name,s) == 0)
+					break;
+			}
+			p->flags |= flags[i].mask;
+			break;
+		case C_ACTION:
+			get_string(s, 32, f);
+			for(i = 0; actions[i].name; i++) {
+				if (strcmp(actions[i].name, s) == 0)
+					break;
+			}
+			p->action = actions[i].act ? : OFF;
+			break;
+		default:
+			free(p);
+			oops_error = -1;
+			return NULL;
+	} while( cmd != C_EOR);
+
+	return p;
+}
+
+/*
+ *	Read the complete state info from the state pipe.
+ *	Returns 0 on success
+ */
+int receive_state(int fd)
+{
+	FILE	*f;
+	char	old_version[256];
+	CHILD	**pp;
+
+	f = fdopen(fd, "r");
+
+ 	if (get_cmd(f) != C_VER)
+		return -1;
+	get_string(old_version, 256, f);
+	oops_error = 0;
+	for (pp = &family; (*pp = get_record(f)) != NULL; pp = &((*pp)->next))
+		;
+	fclose(f);
+	return oops_error;
+}
+
+/*
+ *	Set the process title. We do not check for overflow of
+ *	the stack space since we know there is plenty for
+ *	our needs and we'll never use more than 10 bytes anyway.
+ */
+int setproctitle(char *fmt, ...)
+{
+	va_list ap;
+	int len;
+	char buf[256];
+
+	buf[0] = 0;
+
+	va_start(ap, fmt);
+	len = vsprintf(buf, fmt, ap);
+	va_end(ap);
+
+	memset(argv0, 0, maxproclen + 1);
+	strncpy(argv0, buf, maxproclen);
+
+	return len;
+}
+
+/*
+ *	Open a device with retries.
+ */
+int serial_open(char *dev, int mode)
+{
+	int f, fd;
+	int m;
+
+	/*
+	 *	Open device in nonblocking mode.
+	 */
+	m = mode | O_NONBLOCK;
+
+	/*
+	 *	Retry the open five times.
+	 */
+	for(f = 0;f < 5; f++)
+		if ((fd = open(dev, m)) >= 0) break;
+	if (fd < 0) return(fd);
+
+	/*
+	 *	Set original flags.
+	 */
+	if (m != mode)
+  		fcntl(fd, F_SETFL, mode);
+	return(fd);
+}
+
+/*
+ *	We got a signal (HUP PWR WINCH ALRM INT)
+ */
+void signal_handler(int sig)
+{
+	ADDSET(got_signals, sig);
+}
+
+/*
+ *	SIGCHLD: one of our children has died.
+ */
+void chld_handler()
+{
+	CHILD *ch;
+	int pid, st;
+
+	/*
+	 *	Find out which process(es) this was (were)
+	 */
+	while((pid = waitpid(-1, &st, WNOHANG)) != 0) {
+		if (errno == ECHILD) break;
+		for( ch = family; ch; ch = ch->next )
+			if ( ch->pid == pid && (ch->flags & RUNNING) ) {
+#if DEBUG
+				log(L_VB, "chld_handler: marked %d as zombie",
+					ch->pid);
+#endif
+				ADDSET(got_signals, SIGCHLD);
+				ch->exstat = st;
+				ch->flags |= ZOMBIE;
+				if (ch->new) {
+					ch->new->exstat = st;
+					ch->new->flags |= ZOMBIE;
+				}
+				break;
+			}
+#if DEBUG
+		if (ch == NULL)
+			log(L_VB, "chld_handler: unknown child %d exited.",
+				pid);
+#endif
+	}
+}
+
+/*
+ *	Linux ignores all signals sent to init when the
+ *	SIG_DFL handler is installed. Therefore we must catch SIGTSTP
+ *	and SIGCONT, or else they won't work....
+ *
+ *	The SIGCONT handler
+ */
+void cont_handler()
+{
+	got_cont = 1;
+}
+
+/*
+ *	OOPS: segmentation violation!
+ */
+void segv_handler()
+{
+	log(L_VB, "PANIC: segmentation violation! giving up..");
+	while(1) pause();
+}
+
+/*
+ *	The SIGSTOP & SIGTSTP handler
+ */
+void stop_handler()
+{
+	got_cont = 0;
+	while(!got_cont) pause();
+	got_cont = 0;
+}
+
+/*
+ *	Set terminal settings to reasonable defaults
+ */
+void set_term(int how)
+{
+	struct termios tty;
+	int fd, fd2;
+	int restore_ok = 0;
+
+	if ((fd = serial_open(console_dev, O_RDWR|O_NOCTTY)) < 0) {
+		log(L_VB, "can't open %s", console_dev);
+		return;
+	}
+
+	/*
+	 *	Do we want to save or restore modes.
+	 */
+	if (how > 0) {
+		/*
+		 *	Save the terminal settings.
+		 */
+		(void) ioctl(fd, TCGETS, &tty);
+		if ((fd2 = open(IOSAVE, O_WRONLY|O_CREAT|O_TRUNC, 0600)) < 0) {
+			if (errno != EROFS)
+				log(L_VB, "can't open(%s, O_WRONLY): %s",
+					IOSAVE, strerror(errno));
+			(void) close(fd);
+			(void) close(fd2);
+			return;
+		}
+		(void) write(fd2, &tty, sizeof(struct termios));
+		wrote_iosave = 1;
+		(void) close(fd);
+		(void) close(fd2);
+		return;
+	}
+
+	/*
+	 *	Restore the terminal settings.
+	 */
+	if (wrote_iosave && (fd2 = open(IOSAVE, O_RDONLY)) >= 0) {
+		if (read(fd2, &tty, sizeof(tty)) == sizeof(tty))
+			restore_ok = 1;
+		close(fd2);
+	}
+	if (restore_ok == 0) {
+		/*
+		 *	No old settings - create some resonable defaults.
+		 */
+		(void) ioctl(fd, TCGETS, &tty);
+
+		tty.c_cflag &= CBAUD|CBAUDEX|CSIZE|CSTOPB|PARENB|PARODD;
+		tty.c_cflag |= HUPCL|CLOCAL;
+
+		tty.c_cc[VINTR]  = 3;
+		tty.c_cc[VQUIT]  = 28;
+		tty.c_cc[VERASE] = 127;
+		tty.c_cc[VKILL]  = 24;
+		tty.c_cc[VEOF]   = 4;
+		tty.c_cc[VTIME]  = 0;
+		tty.c_cc[VMIN]   = 1;
+		tty.c_cc[VSTART] = 17;
+		tty.c_cc[VSTOP]  = 19;
+		tty.c_cc[VSUSP]  = 26;
+	}
+
+	/*
+	 *	Set pre and post processing
+	 */
+	tty.c_iflag = IGNPAR|ICRNL|IXON|IXANY;
+	tty.c_oflag = OPOST|ONLCR;
+	tty.c_lflag = ISIG|ICANON|ECHO|ECHOCTL|ECHOPRT|ECHOKE;
+
+	/*
+	 *	Now set the terminal line.
+	 */
+	(void) ioctl(fd, TCSETS, &tty);
+	(void) close(fd);
+}
+
+/*
+ *	Print to the system console
+ */
+void print(char *s)
+{
+	int fd;
+
+	if ((fd = serial_open(console_dev, O_WRONLY|O_NOCTTY|O_NDELAY)) >= 0) {
+		write(fd, s, strlen(s));
+		close(fd);
+	}
+}
+
 /*
  *	Log something to a logfile and the console.
  */
 void log(int loglevel, char *s, ...)
 {
-  va_list va_alist;
-  char buf[256];
+	va_list va_alist;
+	char buf[256];
 
-  va_start(va_alist, s);
-  vsprintf(buf, s, va_alist);
-  va_end(va_alist);
+	va_start(va_alist, s);
+	vsprintf(buf, s, va_alist);
+	va_end(va_alist);
 
-  if (loglevel & L_SY) {
-	/* Re-etablish connection with syslogd every time. */
-	openlog("init", 0, LOG_DAEMON);
-	syslog(LOG_INFO, buf);
-	/* closelog(); NOT needed with recent libc's. */
-  }
+	if (loglevel & L_SY) {
+		/*
+		 *	Re-etablish connection with syslogd every time.
+		 */
+		openlog("init", 0, LOG_DAEMON);
+		syslog(LOG_INFO, buf);
+		/* closelog(); NOT needed with recent libc's. */
+	}
 
-  /* And log to the console. */
-  if (loglevel & L_CO) {
-	print("\rINIT: ");
-	print(buf);
-	print("\r\n");
-  }
+	/*
+	 *	And log to the console.
+	 */
+	if (loglevel & L_CO) {
+		print("\rINIT: ");
+		print(buf);
+		print("\r\n");
+	}
 }
 
 
@@ -582,7 +869,7 @@ int spawn(CHILD *ch, int *res)
 			setsid();
 		}
   		/* Reset all the signals */
-  		for(f = 1; f < _NSIG; f++) signal(f, SIG_DFL);
+  		for(f = 1; f < NSIG; f++) signal(f, SIG_DFL);
   		execvp(args[1], args + 1);
 		/*
 		 *	Is this a bug in execvp? It does _not_ execute shell
@@ -617,51 +904,67 @@ int spawn(CHILD *ch, int *res)
   }
 }
 
-/* Start a child running! */
+/*
+ *	Start a child running!
+ */
 void startup(CHILD *ch)
 {
-  /* See if it's disabled */
-  if (ch->flags & FAILING) return;
+	/*
+	 *	See if it's disabled
+	 */
+	if (ch->flags & FAILING) return;
 
-  switch(ch->action) {
+	switch(ch->action) {
 
-	case SYSINIT:
-  	case BOOTWAIT:
-	case WAIT:
-	case POWERWAIT:
-	case POWERFAILNOW:
-	case POWEROKWAIT:
-	case CTRLALTDEL:
-		if (!(ch->flags & XECUTED)) ch->flags |= WAITING;
-	case KBREQUEST:
-	case BOOT:
-	case POWERFAIL:
-	case ONCE:
-		if (ch->flags & XECUTED) break;
-	case ONDEMAND:
-	case RESPAWN:
-  		ch->flags |= RUNNING;
-  		if (spawn(ch, &(ch->pid)) < 0) break;
-		/* Do NOT log if process field starts with '+' */
-  		if (ch->process[0] != '+')
-			write_utmp_wtmp("", ch->id, ch->pid, INIT_PROCESS, "");
-  		break;
-  }
+		case SYSINIT:
+		case BOOTWAIT:
+		case WAIT:
+		case POWERWAIT:
+		case POWERFAILNOW:
+		case POWEROKWAIT:
+		case CTRLALTDEL:
+			if (!(ch->flags & XECUTED)) ch->flags |= WAITING;
+		case KBREQUEST:
+		case BOOT:
+		case POWERFAIL:
+		case ONCE:
+			if (ch->flags & XECUTED) break;
+		case ONDEMAND:
+		case RESPAWN:
+  			ch->flags |= RUNNING;
+  			if (spawn(ch, &(ch->pid)) < 0) break;
+			/*
+			 *	Do NOT log if process field starts with '+'
+			 *	FIXME: that's for compatibility with *very*
+			 *	old getties - probably it can be taken out.
+			 */
+  			if (ch->process[0] != '+')
+				write_utmp_wtmp("", ch->id, ch->pid,
+					INIT_PROCESS, "");
+  			break;
+	}
 }
 
-/* My version of strtok(3) */
+/*
+ *	My version of strtok(3).
+ */
 char *get_part(char *str, int tok)
 {
-  static char *s;
-  char *p, *q;
+	static char *s;
+	char *p, *q;
 
-  if (str != NULL) s = str;
-  if (s == NULL || *s == 0) return(NULL);
-  q = p = s;
-  while(*p != tok && *p) p++;
-  if (*p == tok) *p++ = 0;
-  s = p;
-  return(q);
+	if (str != NULL)
+		s = str;
+	if (s == NULL || *s == 0)
+		return(NULL);
+	q = p = s;
+	while(*p != tok && *p)
+		p++;
+	if (*p == tok)
+		*p++ = 0;
+	s = p;
+
+	return q;
 }
 
 /*
@@ -746,11 +1049,14 @@ void read_inittab(void)
 	if (!action || !*action)
 			strcpy(err, "missing action field");
 	if (id && strlen(id) > sizeof(utproto.ut_id))
-			sprintf(err, "id field too long (max %d characters)",
-				(int)sizeof(utproto.ut_id));
+		sprintf(err, "id field too long (max %d characters)",
+			(int)sizeof(utproto.ut_id));
 	if (rlevel && strlen(rlevel) > 11)
-			strcpy(err, "rlevel field too long (max 11 characters)");
-	if (process && strlen(process) > 127) strcpy(err, "process field too long");
+		strcpy(err, "rlevel field too long (max 11 characters)");
+	if (process && strlen(process) > 127)
+		strcpy(err, "process field too long");
+	if (action && strlen(action) > 32)
+		strcpy(err, "action field too long");
 	if (err[0] != 0) {
 		log(L_VB, "%s[%d]: %s", INITTAB, lineNo, err);
 #if DEBUG
@@ -789,9 +1095,9 @@ void read_inittab(void)
 	/*
 	 *	Allocate a CHILD structure
 	 */
-	if ((ch = malloc(sizeof(CHILD))) == NULL) {
+	while ((ch = malloc(sizeof(CHILD))) == NULL) {
 		log(L_VB, "out of memory");
-		continue;
+		do_sleep(5);
 	}
 	memset(ch, 0, sizeof(CHILD));
 
@@ -898,7 +1204,7 @@ void read_inittab(void)
 	/*
 	 *	If the entry has changed, kill it anyway. Note that
 	 *	we do not check ch->process, only the "action" field.
-	 *	This way, you can turn an entry "off" immidiately, but
+	 *	This way, you can turn an entry "off" immediately, but
 	 *	changes in the command line will only become effective
 	 *	after the running version has exited.
 	 */
@@ -1049,9 +1355,9 @@ void read_inittab(void)
 }
 
 /*
- * Walk through the family list and start up children.
- * The entries that do not belong here at all are removed
- * from the list.
+ *	Walk through the family list and start up children.
+ *	The entries that do not belong here at all are removed
+ *	from the list.
  */
 void start_if_needed(void)
 {
@@ -1098,64 +1404,72 @@ void start_if_needed(void)
 }
 
 /*
- * Ask the user on the console for a runlevel
+ *	Ask the user on the console for a runlevel
  */
 int ask_runlevel()
 {
-  int lvl = -1;
-  char buf[8];
-  int fd;
+	int lvl = -1;
+	char buf[8];
+	int fd;
 
+	set_term(0);
+	fd = serial_open(console_dev, O_RDWR|O_NOCTTY);
 
-  set_term(0);
-  fd = serial_open(console_dev, O_RDWR|O_NOCTTY);
+	if (fd < 0) return('S');
 
-  if (fd < 0) return('S');
-
-  while(!strchr("0123456789S", lvl)) {
-  	write(fd, "\nEnter runlevel: ", 17);
-	buf[0] = 0;
-  	read(fd, buf, 8);
-  	if (buf[0] != 0 && (buf[1] == '\r' || buf[1] == '\n')) lvl = buf[0];
-	if (islower(lvl)) lvl = toupper(lvl);
-  }
-  close(fd);
-  return(lvl);
+	while(!strchr("0123456789S", lvl)) {
+  		write(fd, "\nEnter runlevel: ", 17);
+		buf[0] = 0;
+  		read(fd, buf, 8);
+  		if (buf[0] != 0 && (buf[1] == '\r' || buf[1] == '\n'))
+			lvl = buf[0];
+		if (islower(lvl)) lvl = toupper(lvl);
+	}
+	close(fd);
+	return lvl;
 }
 
 /*
- * Search the INITTAB file for the 'initdefault' field, with the default
- * runlevel. If this fails, ask the user to supply a runlevel.
+ *	Search the INITTAB file for the 'initdefault' field, with the default
+ *	runlevel. If this fails, ask the user to supply a runlevel.
  */
 int get_init_default(void)
 {
-  CHILD *ch;
-  int lvl = -1;
-  char *p;
+	CHILD *ch;
+	int lvl = -1;
+	char *p;
 
-  /* Look for initdefault. */
-  for(ch = family; ch; ch = ch->next)
-	if (ch->action == INITDEFAULT) {
-		p = ch->rlevel;
-		while(*p) {
-			if (*p > lvl) lvl = *p;
-			p++;
+	/*
+	 *	Look for initdefault.
+	 */
+	for(ch = family; ch; ch = ch->next)
+		if (ch->action == INITDEFAULT) {
+			p = ch->rlevel;
+			while(*p) {
+				if (*p > lvl) lvl = *p;
+				p++;
+			}
+			break;
+		}
+	/*
+	 *	See if level is valid
+	 */
+	if (lvl > 0) {
+		if (islower(lvl)) lvl = toupper(lvl);
+		if (strchr("0123456789S", lvl) == NULL) {
+			log(L_VB, "Initdefault level '%c' is invalid", lvl);
+			lvl = 0;
 		}
-		break;
 	}
-  /* See if level is valid */
-  if (lvl > 0) {
-	if (islower(lvl)) lvl = toupper(lvl);
-	if (strchr("0123456789S", lvl) == NULL) {
-		log(L_VB, "Initdefault level '%c' is invalid", lvl);
-		lvl = 0;
-	}
-  }
-  /* Ask for runlevel on console if needed. */
-  if (lvl <= 0) lvl = ask_runlevel();
+	/*
+	 *	Ask for runlevel on console if needed.
+	 */
+	if (lvl <= 0) lvl = ask_runlevel();
 
-  /* Log the fact that we have a runlevel now. */
-  return(lvl);
+	/*
+	 *	Log the fact that we have a runlevel now.
+	 */
+	return lvl;
 }
 
 
@@ -1177,11 +1491,6 @@ int read_level(int arg)
 
   if (arg) foo = arg;
 
-#if CLEAN_UTMP
-  /* Clean up utmp file first. */
-  write_utmp_wtmp(NULL, "", -2, 0, NULL);
-#endif
-
 #ifdef INITLVL
   ok = 0;
 
@@ -1210,12 +1519,12 @@ int read_level(int arg)
 #endif /* INITLVL */
 
   if (islower(foo)) foo = toupper(foo);
-  if (ok < 1 || ok > 2 || strchr("QS0123456789ABC", foo) == NULL) {
+  if (ok < 1 || ok > 2 || strchr("QS0123456789ABCU", foo) == NULL) {
  	log(L_VB, "bad runlevel: %c", foo);
   	return(runlevel);
   }
 
-  /* Be verbose 'bout it all */
+  /* Log this action */
   switch(foo) {
 	case 'S':
   		log(L_VB, "Going single user");
@@ -1228,6 +1537,9 @@ int read_level(int arg)
 	case 'C':
 		log(L_SY, "Activating demand-procedures for '%c'", foo);
 		break;
+	case 'U':
+		log(L_SY, "Trying to re-exec init");
+		return 'U';
 	default:
 	  	log(L_VB, "Switching to runlevel: %c", foo);
   }
@@ -1312,32 +1624,153 @@ void fail_cancel(void)
   }
 }
 
-/* Start up powerfail entries. */
+/*
+ *	Start up powerfail entries.
+ */
 void do_power_fail(int pwrstat)
 {
-  CHILD *ch;
+	CHILD *ch;
 
-  /* Tell powerwait & powerfail entries to start up */
-  for(ch = family; ch; ch = ch->next) {
-	if (pwrstat == 'O') {
-		/* The power is good again. */
-		if (ch->action == POWEROKWAIT)
-			ch->flags &= ~XECUTED;
-	} else if (pwrstat == 'L') {
-		/* Low battery, shut down now. */
-		if (ch->action == POWERFAILNOW)
-			ch->flags &= ~XECUTED;
-	} else {
-		/* Power is failing */
-		if (ch->action == POWERFAIL || ch->action == POWERWAIT)
-			ch->flags &= ~XECUTED;
+	/*
+	 *	Tell powerwait & powerfail entries to start up
+	 */
+	for (ch = family; ch; ch = ch->next) {
+		if (pwrstat == 'O') {
+			/*
+		 	 *	The power is OK again.
+		 	 */
+			if (ch->action == POWEROKWAIT)
+				ch->flags &= ~XECUTED;
+		} else if (pwrstat == 'L') {
+			/*
+			 *	Low battery, shut down now.
+			 */
+			if (ch->action == POWERFAILNOW)
+				ch->flags &= ~XECUTED;
+		} else {
+			/*
+			 *	Power is failing, shutdown imminent
+			 */
+			if (ch->action == POWERFAIL || ch->action == POWERWAIT)
+				ch->flags &= ~XECUTED;
+		}
 	}
-  }
 }
 
 /*
- * We got a change runlevel request through the
- * init.fifo. Process it.
+ *	Check for state-pipe presence
+ */
+int check_pipe(int fd)
+{
+	struct timeval t;
+	fd_set s;
+	char signature[8];
+
+	FD_ZERO(&s);
+	FD_SET(fd, &s);
+	t.tv_sec = t.tv_usec = 0;
+
+	if (select(fd+1, &s, NULL, NULL, &t) != 1)
+		return 0;
+	if (read(fd, signature, 8) != 8)
+		 return 0;
+	return strcmp(Signature, signature) == 0;
+}
+
+/*
+ *	 Make a state-pipe.
+ */
+int make_pipe(int fd)
+{
+	int fds[2];
+
+	pipe(fds);
+	dup2(fds[0], fd);
+	close(fds[0]);
+	fcntl(fds[1], F_SETFD, 1);
+	fcntl(fd, F_SETFD, 0);
+	write(fds[1], Signature, 8);
+
+	return fds[1];
+}
+
+/*
+ *	Attempt to re-exec.
+ */
+void re_exec(void)
+{
+	sigset_t mask, oldset;
+	pid_t	pid;
+	int	fd;
+	CHILD	*ch;
+
+	if (strchr("S12345",runlevel) == NULL)
+		return;
+	closelog();			/* AAAARRRGH! It _is_ needed */
+
+	/*
+	 *	Reset the alarm, and block all signals.
+	 */
+	alarm(0);
+	sigfillset(&mask);
+	sigprocmask(SIG_BLOCK, &mask, &oldset);
+
+	/*
+	 *	construct a pipe fd --> STATE_PIPE and write a signature
+	 */
+	fd = make_pipe(STATE_PIPE);
+
+	/* 
+	 * It's a backup day today, so I'm pissed off.  Being a BOFH, however, 
+	 * does have it's advantages...
+	 */
+	fail_cancel();
+	close(pipe_fd);
+	pipe_fd = -1;
+	DELSET(got_signals, SIGCHLD);
+	DELSET(got_signals, SIGHUP);
+	DELSET(got_signals, SIGUSR1);
+
+	/*
+	 *	That should be cleaned.
+	 */
+	for(ch = family; ch; ch = ch->next)
+	    if (ch->flags & ZOMBIE) {
+#if DEBUG
+		log(L_VB, "Child died, PID= %d", ch->pid);
+#endif
+		ch->flags &= ~(RUNNING|ZOMBIE|WAITING);
+		if (ch->process[0] != '+')
+			write_utmp_wtmp("", ch->id, ch->pid, DEAD_PROCESS, NULL);
+	    }
+
+	if ((pid = fork()) > 0) {
+		/*
+		 *	Yup. _Parent_ exec's ...
+		 */
+		execl(myname, myname, NULL);
+	} else if (pid == 0) {
+		/*
+		 *	... while child sends her the
+		 *	state information and dies
+		 */
+		send_state(fd);
+		exit(0);
+	}
+	/*
+	 *	We shouldn't be here, something failed. 
+	 *	Bitch, close the state pipe, unblock signals and return.
+	 */
+	close(fd);
+	close(STATE_PIPE);
+	sigprocmask(SIG_SETMASK, &oldset, NULL);
+	log(L_CO, "Attempt to re-exec failed");
+}
+
+
+/*
+ *	We got a change runlevel request through the
+ *	init.fifo. Process it.
  */
 void fifo_new_level(int level)
 {
@@ -1358,20 +1791,28 @@ void fifo_new_level(int level)
 	/* We need to go into a new runlevel */
 	oldlevel = runlevel;
 	runlevel = read_level(level);
-	if (oldlevel != 'S' && runlevel == 'S') set_term(0);
-	if (runlevel == '6' || runlevel == '0' || runlevel == '1') set_term(0);
-	read_inittab();
-	fail_cancel();
-	setproctitle("init [%c]", runlevel);
+	if (runlevel == 'U') {
+		runlevel = oldlevel;
+		re_exec();
+	} else {
+		if (oldlevel != 'S' && runlevel == 'S') set_term(0);
+		if (runlevel == '6' || runlevel == '0' || runlevel == '1') set_term(0);
+		read_inittab();
+		fail_cancel();
+		setproctitle("init [%c]", runlevel);
+	}
   }
 }
 
 /*
- * Read from the init FIFO. Processes like telnetd and rlogind can
- * ask us to create login processes on their behalf.
+ *	Read from the init FIFO. Processes like telnetd and rlogind can
+ *	ask us to create login processes on their behalf.
  *
- * FIXME: this needs to be finished. NOT that it is buggy, but we need
- *	  to add the telnetd/rlogind stuff so people can start using it.
+ *	FIXME:	this needs to be finished. NOT that it is buggy, but we need
+ *		to add the telnetd/rlogind stuff so people can start using it.
+ *		Maybe move to using an AF_UNIX socket so we can use
+ *		the 2.2 kernel credential stuff to see who we're talking to.
+ *	
  */
 void check_init_fifo(void)
 {
@@ -1379,10 +1820,19 @@ void check_init_fifo(void)
   int n;
   fd_set fds;
   int quit = 0;
+  struct stat st;
 
   /* Try to open /dev/initctl */
   if (pipe_fd < 0) {
 	if ((pipe_fd = open(INIT_FIFO, O_RDWR|O_NONBLOCK)) >= 0) {
+		fstat(pipe_fd, &st);
+		if (!S_ISFIFO(st.st_mode)) {
+			log(L_VB, "%s is not a fifo", INIT_FIFO);
+			close(pipe_fd);
+			pipe_fd = -1;
+		}
+	}
+	if (pipe_fd >= 0) {
 		/* Don't use fd's 0, 1 or 2. */
 		(void) dup2(pipe_fd, PIPE_FD);
 		close(pipe_fd);
@@ -1459,16 +1909,15 @@ void check_init_fifo(void)
 
 
 /*
- * This function is used in the transition
- * sysinit (-> single user) boot -> multi-user.
+ *	This function is used in the transition
+ *	sysinit (-> single user) boot -> multi-user.
  */
 void boot_transitions()
 {
   CHILD *ch;
-  static int did_boot = 0;
   static int newlevel = 0;
-  static int loglevel;
-  static int oldlevel;
+  int loglevel;
+  int oldlevel;
   static int warn = 1;
 
   /* Check if there is something to wait for! */
@@ -1555,8 +2004,8 @@ void boot_transitions()
 }
 
 /*
- * Init got hit by a signal. See which signal it is,
- * and act accordingly.
+ *	Init got hit by a signal. See which signal it is,
+ *	and act accordingly.
  */
 void process_signals()
 {
@@ -1578,8 +2027,8 @@ void process_signals()
 		pwrstat = c;
 		close(fd);
 		unlink(PWRSTAT);
-		do_power_fail(pwrstat);
 	}
+	do_power_fail(pwrstat);
 	DELSET(got_signals, SIGPWR);
   }
 
@@ -1649,12 +2098,17 @@ void process_signals()
 #ifdef INITLVL
 		runlevel = read_level(0);
 #endif
-		if (oldlevel != 'S' && runlevel == 'S') set_term(0);
-		if (runlevel == '6' || runlevel == '0' || runlevel == '1') set_term(0);
-		read_inittab();
-		fail_cancel();
-		setproctitle("init [%c]", runlevel);
-		DELSET(got_signals, SIGHUP);
+		if (runlevel == 'U') {
+			runlevel = oldlevel;
+			re_exec();
+		} else {
+			if (oldlevel != 'S' && runlevel == 'S') set_term(0);
+			if (runlevel == '6' || runlevel == '0' || runlevel == '1') set_term(0);
+			read_inittab();
+			fail_cancel();
+			setproctitle("init [%c]", runlevel);
+			DELSET(got_signals, SIGHUP);
+		}
 	}
   }
   if (ISMEMBER(got_signals, SIGUSR1)) {
@@ -1666,50 +2120,57 @@ void process_signals()
 #endif
 	close(pipe_fd);
 	pipe_fd = -1;
+	DELSET(got_signals, SIGUSR1);
   }
 }
 
 /*
- * The main loop
+ *	The main loop
  */ 
 int init_main()
 {
   int f, st;
   CHILD *ch;
+  sigset_t sgt;
   struct sigaction sa;
   char *s;
 
+  if (!reload) {
+  
 #if INITDEBUG
-  /*
-   * Fork so we can debug the init process.
-   */
-  if ((f = fork()) > 0) {
-	while(wait(&st) != f)
-		;
-	write(1, "PRNT: init killed.\r\n", 20);
-	while(1) pause();
+	/*
+	 * Fork so we can debug the init process.
+	 */
+	if ((f = fork()) > 0) {
+		while(wait(&st) != f)
+			;
+		write(1, "PRNT: init killed.\r\n", 20);
+		while(1) pause();
+	}
+#endif
+
+	/*
+	 *	Tell the kernel to send us SIGINT when CTRL-ALT-DEL
+	 *	is pressed, and that we want to handle keyboard signals.
+	 */
+	init_reboot(BMAGIC_SOFT);
+	if ((f = open(VT_MASTER, O_RDWR | O_NOCTTY)) >= 0) {
+		(void) ioctl(f, KDSIGACCEPT, SIGWINCH);
+		close(f);
+	} else
+		(void) ioctl(0, KDSIGACCEPT, SIGWINCH);
+
+	/*
+	 *	Ignore all signals.
+	 */
+	for(f = 1; f <= NSIG; f++)
+		signal(f, SIG_IGN);
   }
-#endif
-
-  /* Tell the kernel to send us SIGINT when CTRL-ALT-DEL is pressed */
-  init_reboot(BMAGIC_SOFT);
-
-  /* Tell the kernel that we want to handle keyboard signals. */
-#ifdef KDSIGACCEPT
-  if ((f = open(VT_MASTER, O_RDWR | O_NOCTTY)) >= 0) {
-  	(void) ioctl(f, KDSIGACCEPT, SIGWINCH);
-	close(f);
-  } else
-  	(void) ioctl(0, KDSIGACCEPT, SIGWINCH);
-#endif
-
-  /* Set up signals */
-  for(f = 1; f <= _NSIG; f++)
-  	signal(f, SIG_IGN);
 
   SETSIG(sa, SIGALRM,  signal_handler);
   SETSIG(sa, SIGHUP,   signal_handler);
   SETSIG(sa, SIGINT,   signal_handler);
+  SETSIG(sa, SIGCHLD,  chld_handler);
   SETSIG(sa, SIGPWR,   signal_handler);
   SETSIG(sa, SIGWINCH, signal_handler);
   SETSIG(sa, SIGUSR1,  signal_handler);
@@ -1730,37 +2191,59 @@ int init_main()
   siginterrupt (SIGUSR1,  1);
   siginterrupt (SIGWINCH, 1);
 
-  /* Close whatever files are open, and reset the console. */
-  close(0);
-  close(1);
-  close(2);
-  if ((s = getenv("CONSOLE")) != NULL)
-	console_dev = s;
-  set_term(0);
-  setsid();
+  if (!reload) {
 
-  /* Set default PATH variable (for ksh) */
-  if (getenv("PATH") == NULL) putenv(PATH_DFL);
+  	/* Close whatever files are open, and reset the console. */
+	close(0);
+	close(1);
+	close(2);
+	if ((s = getenv("CONSOLE")) != NULL)
+		console_dev = s;
+  	set_term(0);
+  	setsid();
 
-  /* Initialize /var/run/utmp (only works if /var is on root and mounted rw) */
-  (void) close(open(UTMP_FILE, O_WRONLY|O_CREAT|O_TRUNC, 0644));
+  	/*
+	 *	Set default PATH variable (for ksh)
+	 */
+  	if (getenv("PATH") == NULL) putenv(PATH_DFL);
 
-  /* Say hello to the world */
-  log(L_CO, bootmsg);
+  	/*
+	 *	Initialize /var/run/utmp (only works if /var is on
+	 *	root and mounted rw)
+	 */
+  	(void) close(open(UTMP_FILE, O_WRONLY|O_CREAT|O_TRUNC, 0644));
 
-  /* See if we have to start an emergency shell. */
-  if (emerg_shell) {
-	SETSIG(sa, SIGCHLD, SIG_DFL);
-	if (spawn(&ch_emerg, &f) > 0) {
-		while(wait(&st) != f)
-			;
-	}
+  	/*
+	 *	Say hello to the world
+	 */
+  	log(L_CO, bootmsg);
+
+  	/*
+	 *	See if we have to start an emergency shell.
+	 */
+	if (emerg_shell) {
+		SETSIG(sa, SIGCHLD, SIG_DFL);
+		if (spawn(&ch_emerg, &f) > 0) {
+			while(wait(&st) != f)
+				;
+		}
+  		SETSIG(sa, SIGCHLD,  chld_handler);
+  	}
+
+  	/*
+	 *	Start normal boot procedure.
+	 */
+  	runlevel = '#';
+  	read_inittab();
+  
+  } else {
+	/*
+	 *	Restart: unblock signals and let the show go on
+	 */
+	log(L_CO, bootmsg);
+	sigfillset(&sgt);
+	sigprocmask(SIG_UNBLOCK, &sgt, NULL);
   }
-  SETSIG(sa, SIGCHLD,  chld_handler);
-
-  /* Start normal boot procedure. */
-  runlevel = '#';
-  read_inittab();
   start_if_needed();
 
   while(1) {
@@ -1774,10 +2257,6 @@ int init_main()
      /* Check if there are processes to be waited on. */
      for(ch = family; ch; ch = ch->next)
 	if ((ch->flags & RUNNING) && ch->action != BOOT) break;
-#if 0
-     if (ch == NULL && got_signals == 0)
-	log(L_VB, "no more processes in this runlevel");
-#endif
 
 #if CHANGE_WAIT
      /* Wait until we get hit by some signal. */
@@ -1801,13 +2280,6 @@ int init_main()
 
      /* See what we need to start up (again) */
      start_if_needed();
-#if 0
-     /*
-      *	Sync'ing in the main loop is a bad idea they say,
-      *	so it has been disabled for now.
-      */
-     sync();
-#endif
   }
   /*NOTREACHED*/
 }
@@ -1817,10 +2289,11 @@ int init_main()
  */
 void Usage(char *s)
 {
-  fprintf(stderr, "Usage: %s 0123456SsQqAaBbCc\n", s);
+  fprintf(stderr, "Usage: %s 0123456SsQqAaBbCcUu\n", s);
   exit(1);
 }
 
+
 /*
  * Main entry for init and telinit.
  */
@@ -1838,12 +2311,29 @@ int main(int argc, char *argv[])
   	p = argv[0];
   	
 
-  /* See if I want to become "parental guide of all processes" */
+  /*
+   *	Is this telinit or init ?
+   */
 #ifndef TEST
-  if (getpid() == INITPID || strcmp(p, "init.new") == 0 || strcmp(p, "sh") == 0) {
-#else
-			   {
+  if (getpid() == INITPID || !strcmp(p, "init.new") || !strcmp(p, "sh"))
 #endif
+  {
+	/*
+	 *	Check for re-exec
+	 */ 	
+	if (check_pipe(STATE_PIPE)) {
+
+		receive_state(STATE_PIPE);
+
+		myname = strdup(argv[0]);
+		argv0 = argv[0];
+		maxproclen = strlen(argv[0]);
+		reload = 1;
+		wrote_iosave = 1;
+
+		init_main();
+	}
+
   	/* Check command line arguments */
 	maxproclen = strlen(argv[0]) + 1;
   	for(f = 1; f < argc; f++) {
@@ -1856,6 +2346,7 @@ int main(int argc, char *argv[])
 		else if (strchr("0123456789sS", argv[f][0])
 			&& strlen(argv[f]) == 1)
 			dfl_level = argv[f][0];
+		/* "init u" in the very beginning makes no sence */
 		if (dfl_level == 's') dfl_level = 'S';
 		maxproclen += strlen(argv[f]) + 1;
 	}
@@ -1887,7 +2378,7 @@ int main(int argc, char *argv[])
 
   /* Check syntax. */
   if (argc - optind != 1 || strlen(argv[optind]) != 1) Usage(p);
-  if (!strchr("0123456789SsQqAaBbCc", argv[optind][0])) Usage(p);
+  if (!strchr("0123456789SsQqAaBbCcUu", argv[optind][0])) Usage(p);
 
   umask(022);
 
