@@ -13,10 +13,10 @@
  *
  * Author:	Miquel van Smoorenburg, miquels@cistron.nl
  *
- * Version:	@(#)shutdown  2.73  25-Nov-1997  MvS
+ * Version:	@(#)shutdown  2.74  10-May-1998  miquels@cistron.nl
  *
  *		This file is part of the sysvinit suite,
- *		Copyright 1991-1997 Miquel van Smoorenburg.
+ *		Copyright 1991-1998 Miquel van Smoorenburg.
  *
  *		This program is free software; you can redistribute it and/or
  *		modify it under the terms of the GNU General Public License
@@ -37,7 +37,7 @@
 #include "paths.h"
 #include "reboot.h"
 
-char *Version = "@(#) shutdown 2.73 25-Nov-1997 MvS";
+char *Version = "@(#) shutdown 2.74 10-May-1998 miquels@cistron.nl";
 
 #define MESSAGELEN	256
 
@@ -90,7 +90,8 @@ void stopit()
 void usage()
 {
  fprintf(stderr,
-	"Usage:\t  shutdown [-krhfnc] [-t secs] time [warning message]\n"
+	"Usage:\t  shutdown [-akrhfnc] [-t secs] time [warning message]\n"
+	"\t\t  -a:      use /etc/shutdown.allow\n"
 	"\t\t  -k:      don't really shutdown, only warn.\n"
 	"\t\t  -r:      reboot after shutdown.\n"
 	"\t\t  -h:      halt after shutdown.\n"
@@ -282,7 +283,9 @@ char **argv;
   char *when = NULL;
   int didnolog = 0;
   int cancel = 0;
+  int useacl = 0;
   int pid = 0;
+  uid_t realuid;
   FILE *fp;
   char *downusers[32];
   char buf[128];
@@ -292,6 +295,7 @@ char **argv;
   struct sigaction sa;
 
   /* We can be installed setuid root (executable for a special group) */
+  realuid = getuid();
   setuid(geteuid());
 
   if (getuid() != 0) {
@@ -300,59 +304,12 @@ char **argv;
   }
   strcpy(down_level, "1");
 
-  /* See if we have a controlling terminal. */
-  if (tcgetpgrp(0) != getpgrp() && getppid() == 1 &&
-      (fp = fopen(SDALLOW, "r")) != NULL) {
-
-	/* No controlling terminal, we were called from init. */
-	i = 0;
-	/* Read /etc/shutdown.allow. */
-	while(fgets(buf, 128, fp)) {
-		if (buf[0] == '#' || buf[0] == '\n') continue;
-		if (i > 31) continue;
-		for(sp = buf; *sp; sp++) if (*sp == '\n') *sp = 0;
-		downusers[i++] = strdup(buf);
-	}
-	if (i < 32) downusers[i] = 0;
-	fclose(fp);
-
-	/* Now walk through /var/run/utmp to find logged in users. */
-	while(!user_ok && (ut = getutent()) != NULL) {
-
-		/* See if this is a user process on a VC. */
-		if (ut->ut_type != USER_PROCESS) continue;
-		sprintf(buf, "/dev/%s", ut->ut_line);
-		if (stat(buf, &st) < 0) continue;
-		if ((st.st_rdev & 0xFFC0) != 0x0400) continue;
-
-		/* Root is always OK. */
-		if (strcmp(ut->ut_user, "root") == 0) {
-			user_ok++;
-			break;
-		}
-
-		/* See if this is an allowed user. */
-		for(i = 0; i < 32 && downusers[i]; i++)
-			if (!strncmp(downusers[i], ut->ut_user, UT_NAMESIZE)) {
-				user_ok++;
-				break;
-			}
-	}
-	endutent();
-
-	/* See if user was allowed. */
-	if (!user_ok) {
-		if ((fp = fopen(CONSOLE, "w")) != NULL) {
-		   fprintf(fp, "\rshutdown: no authorized users logged in.\r\n");
-		   fclose(fp);
-		}
-		exit(1);
-	}
-  }
-
   /* Process the options. */
-  while((c = getopt(argc, argv, "cqkrhnfFyt:g:i:")) != EOF) {
+  while((c = getopt(argc, argv, "acqkrhnfFyt:g:i:")) != EOF) {
   	switch(c) {
+		case 'a': /* Access control. */
+			useacl = 1;
+			break;
 		case 'c': /* Cancel an already running shutdown. */
 			cancel = 1;
 			break;
@@ -396,6 +353,55 @@ char **argv;
   			break;	
   	}
   }
+
+  /* Do we need to use the shutdown.allow file ? */
+  if (useacl && (fp = fopen(SDALLOW, "r")) != NULL) {
+
+	/* Read /etc/shutdown.allow. */
+	i = 0;
+	while(fgets(buf, 128, fp)) {
+		if (buf[0] == '#' || buf[0] == '\n') continue;
+		if (i > 31) continue;
+		for(sp = buf; *sp; sp++) if (*sp == '\n') *sp = 0;
+		downusers[i++] = strdup(buf);
+	}
+	if (i < 32) downusers[i] = 0;
+	fclose(fp);
+
+	/* Now walk through /var/run/utmp to find logged in users. */
+	while(!user_ok && (ut = getutent()) != NULL) {
+
+		/* See if this is a user process on a VC. */
+		if (ut->ut_type != USER_PROCESS) continue;
+		sprintf(buf, "/dev/%s", ut->ut_line);
+		if (stat(buf, &st) < 0) continue;
+		if ((st.st_rdev & 0xFFC0) != 0x0400) continue;
+
+		/* Root is always OK. */
+		if (strcmp(ut->ut_user, "root") == 0) {
+			user_ok++;
+			break;
+		}
+
+		/* See if this is an allowed user. */
+		for(i = 0; i < 32 && downusers[i]; i++)
+			if (!strncmp(downusers[i], ut->ut_user, UT_NAMESIZE)) {
+				user_ok++;
+				break;
+			}
+	}
+	endutent();
+
+	/* See if user was allowed. */
+	if (!user_ok) {
+		if ((fp = fopen(CONSOLE, "w")) != NULL) {
+		   fprintf(fp, "\rshutdown: no authorized users logged in.\r\n");
+		   fclose(fp);
+		}
+		exit(1);
+	}
+  }
+
   /* Read pid of running shutdown from a file */
   if ((fp = fopen(SDPID, "r")) != NULL) {
 	fscanf(fp, "%d", &pid);
