@@ -11,7 +11,7 @@
  *		encrypted root password is "x" the shadow
  *		password will be used.
  *
- * Version:	@(#)sulogin 2.57 26-Nov-1997 MvS.
+ * Version:	@(#)sulogin 2.74 10-May-1998 miquels@cistron.nl
  *
  */
 
@@ -37,7 +37,7 @@
 #define SHADOW	"/etc/shadow"
 #define BINSH	"/bin/sh"
 
-char *Version = "@(#)sulogin 2.57 26-Nov-1997 MvS.";
+char *Version = "@(#)sulogin 2.74 10-May-1998 miquels@cistron.nl";
 
 int timeout = 0;
 
@@ -97,6 +97,7 @@ void alrm_handler()
 int valid(char *pass)
 {
 	char *s;
+	int len;
 
 	if (pass[0] == 0) return 1;
 #if CHECK_MD5
@@ -110,7 +111,8 @@ int valid(char *pass)
 		for(s = pass + 3; *s && *s != '$'; s++)
 			;
 		if (*s++ != '$') return 0;
-		if (strlen(s) != 24) return 0;
+		len = strlen(s);
+		if (len < 22 || len > 24) return 0;
 
 		return 1;
 	}
