@@ -2,7 +2,7 @@
  * utmp.c	Routines to read/write the utmp and wtmp files.
  *		Basically just wrappers around the library routines.
  *
- * Version:	@(#)utmp.c  1.21  12-Oct-1997  miquels@cistron.nl
+ * Version:	@(#)utmp.c  1.22  22-Jan-1998  miquels@cistron.nl
  *
  */
 #include <sys/types.h>
@@ -21,8 +21,13 @@
 #include "initreq.h"
 #include "paths.h"
 
+
 #if defined(__GLIBC__)
-#  define HAVE_UPDWTMP 1
+#  if (__GLIBC__ == 2) && (__GLIBC_MINOR__ == 0) && defined(__powerpc__)
+#    define HAVE_UPDWTMP 0
+#  else
+#    define HAVE_UPDWTMP 1
+#  endif
 #else
 #  define HAVE_UPDWTMP 0
 #endif
